﻿using System;
using System.Collections.Generic;
using System.Text;
// 추가
using System.Collections;
using System.Data;
using FxLib.Utils;

namespace MyOpenBom.Common
{
    /// <summary>
    /// 아래 클래스는 크게 3가지 프로그램적 특징을 가지고 잇다.
    /// 1. Composite 패턴 적용
    /// 2. 컬렉션 객체내의 특정 노드 탐색 시 재귀 함수 호출
    ///     => Find(), DisplayTree()에 해당
    /// 3. 반복기 사용(foreach), 컬렉션 클래스 사용을 위한 yield 키워드 사용
    /// 
    /// 반복기(foreach) 사용은 주로 컬렉션의 횡적 노드 탐색을 주 목적으로 하지만
    /// 종적 재귀 호출에 응용된다.
    /// </summary>
    /// 

    // "Component" Treenode 
    // 반복기 사용(foreach)를 위한 IEnumerable 인터페이스 상속함
    public abstract class CodeComponent : IEnumerable
    {
        protected string code;      // 코드
        protected string name;      // 코드값

        public CodeComponent(string code, string name)
        {
            this.code = code;
            this.name = name;
        }

        // 각 객체의 operation
        public abstract string Code();
        public abstract string Name();

        // 집합객체의 operation
        public abstract void Add(CodeComponent d);
        public abstract void Remove(CodeComponent d);
        public abstract void DisplayTree(int indent);           // 트리 구조를 표시한다.
        public abstract void Clear();
        public abstract CodeComponent Find(string code);        // 특정 코드에 대한 객체를 반환한다.

        public abstract IEnumerator GetEnumerator();            // 반복기 사용을 위함
    }

    // "Leaf" 
    public class CodeElement : CodeComponent, IEnumerable
    {
        // Constructor 
        public CodeElement(string code, string name)
            : base(code, name)
        {
        }

        public override void Add(CodeComponent c)
        {
            Console.WriteLine("Cannot add");
        }

        public override void Remove(CodeComponent c)
        {
            Console.WriteLine("Cannot remove");
        }

        public override void DisplayTree(int indent)
        {
            Console.WriteLine(
              new String('-', indent) + " " + code + " " + name);
        }

        public override void Clear()
        {
            Console.WriteLine("Cannot add");
        }

        public override string Code()
        {
            return code;
        }

        public override string Name()
        {
            return name;
        }

        public override CodeComponent Find(string code)
        {
            return null;
        }

        public override IEnumerator GetEnumerator()
        {
            yield return this;
        }
    }


    // "Composite" 
    public class CodeComposite : CodeComponent, IEnumerable
    {
        private ArrayList elements = new ArrayList();

        // Constructor 
        public CodeComposite(string code, string name)
            : base(code, name)
        {
        }

        public override void Add(CodeComponent d)
        {
            elements.Add(d);
        }

        public override void Remove(CodeComponent d)
        {
            elements.Remove(d);
        }

        public override void DisplayTree(int indent)
        {
            Console.WriteLine(new String('-', indent) +
              "+ " + name);

            // Display each child element on this node 
            foreach (CodeComponent c in elements)
            {
                c.DisplayTree(indent + 2);
            }
        }

        public override void Clear()
        {
            // Display each child element on this node 
            foreach (CodeComponent c in elements)
            {
                c.Clear();
            }
            elements.Clear();
        }

        public override string Code()
        {
            return code;
        }

        public override string Name()
        {
            return name;
        }

        /// <summary>
        /// 주어진 코드에 대한 루트 코드 객체를 반환한다.
        /// 이 함수는 트리내에 모든 노드가 고유함을 가정한 코드이다.
        /// 만약 BOM과 같은 구조에서는 중복 요소가 많을 수 있기 때문에
        /// 아래 메소드를 사용해서는 안된다.
        /// 
        /// 재귀적 호출 사용
        /// </summary>
        /// <param name="code"></param>
        /// <returns></returns>
        public override CodeComponent Find(string code)
        {
            foreach (CodeComponent o in elements)
            {
                // 코드가 일치하는지 확인하고
                if (o.Code().Equals(code))
                    return o;

                // 아니면 개체 내에서 재귀적 호출하고...
                if (o.Find(code) != null)
                    return o;
            }
            return null;
        }

        // 반복기 사용
        public override IEnumerator GetEnumerator()
        {
            foreach (CodeComponent o in elements)
            {
                yield return o;
            }
        }
    }
}
